#include <cmath>
#include <msp/gl/meshbuilder.h>
#include <msp/gl/texture2d.h>
#include "oculusriftcombiner.h"
#include "oculusriftdevice.h"
#include "oculusriftdevice_private.h"

using namespace std;


namespace {

const char vs_source[] =
	"varying vec2 texcoord_r;\n"
	"varying vec2 texcoord_g;\n"
	"varying vec2 texcoord_b;\n"
	"varying float vignette;\n"
	"void main()\n"
	"{\n"
	"	gl_Position = vec4(gl_Vertex.xy, 0.5, 1.0);\n"
	"	texcoord_r = gl_MultiTexCoord0.xy;\n"
	"	texcoord_g = gl_MultiTexCoord1.xy;\n"
	"	texcoord_b = gl_MultiTexCoord2.xy;\n"
	"	vignette = gl_MultiTexCoord3.x;\n"
	"}\n";

const char fs_source[] =
	"uniform sampler2D texture;\n"
	"varying vec2 texcoord_r;\n"
	"varying vec2 texcoord_g;\n"
	"varying vec2 texcoord_b;\n"
	"varying float vignette;\n"
	"void main()\n"
	"{\n"
	"	float r = texture2D(texture, texcoord_r).r;\n"
	"	float g = texture2D(texture, texcoord_g).g;\n"
	"	float b = texture2D(texture, texcoord_b).b;\n"
	"	gl_FragColor = vec4(vec3(r, g, b)*vignette, 1.0);\n"
	"}\n";

void create_distortion_mesh(Msp::GL::Mesh &mesh, ovrHmd hmd, ovrEyeType eye, const ovrFovPort &fov)
{
	ovrDistortionMesh ovr_mesh;
	ovrHmd_CreateDistortionMesh(hmd, eye, fov, ovrDistortionCap_Chromatic|ovrDistortionCap_Vignette, &ovr_mesh);

	ovrSizei tex_size = ovrHmd_GetFovTextureSize(hmd, eye, fov, 1.0);
	ovrRecti view_rect;
	view_rect.Pos.x = 0;
	view_rect.Pos.y = 0;
	view_rect.Size = tex_size;
	ovrVector2f uv_scale_offset[2];
	ovrHmd_GetRenderScaleAndOffset(fov, tex_size, view_rect, uv_scale_offset);
	ovrVector2f &scale = uv_scale_offset[0];
	ovrVector2f &offset = uv_scale_offset[1];

	Msp::GL::MeshBuilder bld(mesh);
	for(unsigned i=0; i<ovr_mesh.VertexCount; ++i)
	{
		ovrDistortionVertex &v = ovr_mesh.pVertexData[i];
		bld.multitexcoord(0, v.TanEyeAnglesR.x*scale.x+offset.x, 1.0f-(v.TanEyeAnglesR.y*scale.y+offset.y));
		bld.multitexcoord(1, v.TanEyeAnglesG.x*scale.x+offset.x, 1.0f-(v.TanEyeAnglesG.y*scale.y+offset.y));
		bld.multitexcoord(2, v.TanEyeAnglesB.x*scale.x+offset.x, 1.0f-(v.TanEyeAnglesB.y*scale.y+offset.y));
		bld.multitexcoord(3, v.VignetteFactor);
		bld.vertex(v.ScreenPosNDC.x, v.ScreenPosNDC.y);
	}

	Msp::GL::Batch batch(Msp::GL::TRIANGLES);
	batch.append(vector<unsigned>(ovr_mesh.pIndexData, ovr_mesh.pIndexData+ovr_mesh.IndexCount));
	mesh.add_batch(batch);

	ovrHmd_DestroyDistortionMesh(&ovr_mesh);
}

}

namespace Msp {
namespace VR {

OculusRiftCombiner::OculusRiftCombiner(const OculusRiftDevice &d):
	device(d),
	left_mesh((GL::VERTEX2, GL::TEXCOORD2,0, GL::TEXCOORD2,1, GL::TEXCOORD2,2, GL::TEXCOORD1,3)),
	right_mesh((GL::VERTEX2, GL::TEXCOORD2,0, GL::TEXCOORD2,1, GL::TEXCOORD2,2, GL::TEXCOORD1,3)),
	shprog(vs_source, fs_source)
{
	const OculusRiftDevice::Private &dev_priv = device.get_private();
	ovrHmd hmd = dev_priv.ovr_hmd;

	ovrFovPort left_fov = hmd->DefaultEyeFov[ovrEye_Left];
	ovrFovPort right_fov = hmd->DefaultEyeFov[ovrEye_Right];
	float vertical = max(max(left_fov.UpTan, left_fov.DownTan), max(right_fov.UpTan, right_fov.DownTan));
	fov = Geometry::atan<float>(vertical)*2.0f;

	float inner = max(left_fov.RightTan, right_fov.LeftTan);
	float outer = max(left_fov.LeftTan, right_fov.RightTan);
	frustum_skew = (inner-outer)/(inner+outer);

	left_fov.UpTan = right_fov.UpTan = vertical;
	left_fov.DownTan = right_fov.DownTan = vertical;
	left_fov.RightTan = right_fov.LeftTan = inner;
	left_fov.LeftTan = right_fov.RightTan = outer;

	create_distortion_mesh(left_mesh, hmd, ovrEye_Left, left_fov);
	create_distortion_mesh(right_mesh, hmd, ovrEye_Right, right_fov);

	ovrSizei tex_size = ovrHmd_GetFovTextureSize(hmd, ovrEye_Left, left_fov, 1.0);
	width_factor = tex_size.w*1.0f/hmd->Resolution.w;
	height_factor = tex_size.h*1.0f/hmd->Resolution.h;
	float aspect = (inner+outer)/(vertical*2);
	aspect_factor = aspect*hmd->Resolution.h/hmd->Resolution.w;

	shdata.uniform("texture", 0);

}

void OculusRiftCombiner::render(const GL::Texture2D &left, const GL::Texture2D &right) const
{
	GL::Bind bind_shprog(shprog);
	shdata.apply();

	GL::Bind bind_tex(left);
	left_mesh.draw();

	right.bind();
	right_mesh.draw();
}

} // namespace VR
} // namespace Msp
